@@ -3,8 +3,7 @@ import random
 import struct
 from hashlib import sha1, sha256, md5
 from time import time
-
-from telethon import utils
+from telethon import utils, hints
 from telethon.crypto import AES
 from telethon.errors import SecurityError, EncryptionAlreadyDeclinedError
 from telethon.extensions import BinaryReader
@@ -13,8 +12,8 @@ from telethon.tl.functions.messages import AcceptEncryptionRequest, SendEncrypte
 from telethon.tl.functions.messages import GetDhConfigRequest, RequestEncryptionRequest, SendEncryptedServiceRequest, \
     DiscardEncryptionRequest, SendEncryptedRequest
 from telethon.tl.types import InputEncryptedChat, TypeEncryptedChat, EncryptedFile, InputEncryptedFileLocation, \
-    InputFileBig, InputFile, InputEncryptedFileBigUploaded, InputEncryptedFileUploaded
-from telethon.tl.types.messages import DhConfigNotModified
+    InputFileBig, InputFile, InputEncryptedFileBigUploaded, InputEncryptedFileUploaded, EncryptedChat
+from telethon.tl.types.messages import DhConfigNotModified, DhConfig, SentEncryptedMessage
 
 from .secret_sechma.secretTL import DecryptedMessageService, DecryptedMessageActionRequestKey, \
     DecryptedMessageActionAcceptKey, DecryptedMessageActionAbortKey, DecryptedMessageActionCommitKey, \
@@ -47,53 +46,58 @@ def _old_calc_key(auth_key, msg_key, client):
     return aes_key, aes_iv
 
 
-class ChatKey:
-    def __init__(self, auth_key: bytes):
-        self.auth_key = auth_key
-        self.fingerprint = None
-
-
-class Chats:
-    def __init__(self, id: int, access_hash: int, key: ChatKey, admin: bool, user_id: int,
-                 input_chat: InputEncryptedChat):
+class SecretChat:
+    def __init__(self, id: int, access_hash: int, auth_key: bytes, admin: bool, user_id: int,
+                 input_chat: [InputEncryptedChat, None], created=time(), updated=time(), in_seq_no_x=None,
+                 out_seq_no_x=None, in_seq_no=0, out_seq_no=0, layer=DEFAULT_LAYER, ttl=0, ttr=100,
+                 mtproto=1):
         self.id = id
         self.access_hash = access_hash
-        self.key = key
+        self.auth_key = auth_key
         self.admin = admin
         self.user_id = user_id
         self.input_chat = input_chat
-        self.in_seq_no_x = 0 if admin else 1
-        self.out_seq_no_x = 1 if admin else 0
-        self.in_seq_no = 0
-        self.out_seq_no = 0
-        self.layer = DEFAULT_LAYER
-        self.ttl = 0
-        self.ttr = 100
-        self.updated = time()
+        if not in_seq_no_x:
+            self.in_seq_no_x = 0 if admin else 1
+        else:
+            self.in_seq_no_x = in_seq_no_x
+        if not out_seq_no_x:
+
+            self.out_seq_no_x = 1 if admin else 0
+        else:
+            self.out_seq_no_x = out_seq_no_x
+        self.in_seq_no = in_seq_no
+        self.out_seq_no = out_seq_no
+        self.layer = layer
+        self.ttl = ttl
+        self.ttr = ttr
+        self.updated = updated
+        # TODO store these maybe too
         self.incoming = {}
         self.outgoing = {}
-        self.created = time()
+        self.created = created
+        # We probably don't need to store these
         self.rekeying = [0]
-        self.mtproto = 1
+        self.mtproto = mtproto
 
 
 class SecretChatMethods:
 
-    def get_secret_chat(self, chat_id) -> Chats:
+    def get_secret_chat(self, chat_id: [int, SecretChat]) -> SecretChat:
         if isinstance(chat_id, int):
-            peer = self.secret_chats.get(chat_id, None)
+            peer = self.session.get_secret_chat_by_id(chat_id)
             if not peer:
                 raise ValueError("chat not found")
             return peer
         try:
-            peer = self.secret_chats.get(chat_id.id, None)
+            peer = self.session.get_secret_chat_by_id(chat_id.id)
             if not peer:
                 raise ValueError("chat not found")
             return peer
         except AttributeError:
             pass
         try:
-            peer = self.secret_chats.get(chat_id.chat_id, None)
+            peer = self.session.get_secret_chat_by_id(chat_id.chat_id)
             if not peer:
                 raise ValueError("chat not found")
             return peer
@@ -104,11 +108,13 @@ class SecretChatMethods:
     async def get_dh_config(self):
         version = 0 if not self.dh_config else self.dh_config.version
         dh_config = await self.client(GetDhConfigRequest(random_length=0, version=version))
+
         if isinstance(dh_config, DhConfigNotModified):
             return self.dh_config
-        dh_config.p = int.from_bytes(dh_config.p, 'big', signed=False)
-        self.dh_config = dh_config
-        return dh_config
+        elif isinstance(dh_config, DhConfig):
+            dh_config.p = int.from_bytes(dh_config.p, 'big', signed=False)
+            self.dh_config = dh_config
+            return dh_config
 
     def check_g_a(self, g_a: int, p: int) -> bool:
         if g_a <= 1 or g_a >= p - 1:
@@ -117,25 +123,28 @@ class SecretChatMethods:
             raise ValueError("g_a is invalid (1 < g_a < p - 1 is false).")
         return True
 
-    async def start_secret_chat(self, peer):
-        peer = utils.get_input_user(await self.get_input_entity(peer))
+    async def start_secret_chat(self, peer: 'hints.EntityLike'):
+        peer = utils.get_input_user(await self.client.get_input_entity(peer))
         dh_config = await self.get_dh_config()
         a = int.from_bytes(os.urandom(256), 'big', signed=False)
         g_a = pow(dh_config.g, a, dh_config.p)
         self.check_g_a(g_a, dh_config.p)
         res = await self.client(RequestEncryptionRequest(user_id=peer, g_a=g_a.to_bytes(256, 'big', signed=False)))
-        self.temp_secret_chat[res.id] = a
+        temp_chat = SecretChat(res.id, 0, a.to_bytes(256, 'big', signed=False), False, 0, None)
+        self.session.save_chat(temp_chat, True)
         return res.id
 
     def generate_secret_in_seq_no(self, chat_id):
-        return self.secret_chats[chat_id].in_seq_no * 2 + self.secret_chats[chat_id].in_seq_no_x
+        chat = self.session.get_secret_chat_by_id(chat_id)
+        return chat.in_seq_no * 2 + chat.in_seq_no_x
 
     def generate_secret_out_seq_no(self, chat_id):
-        return self.secret_chats[chat_id].out_seq_no * 2 + self.secret_chats[chat_id].out_seq_no_x
+        chat = self.session.get_secret_chat_by_id(chat_id)
+        return chat.out_seq_no * 2 + chat.out_seq_no_x
 
     async def rekey(self, peer):
         peer = self.get_secret_chat(peer)
-        self._log.client[__name__].debug(f'Rekeying secret chat {peer}')
+        self._log.debug(f'Rekeying secret chat {peer}')
         dh_config = await self.get_dh_config()
         a = int.from_bytes(os.urandom(256), 'big', signed=False)
         g_a = pow(dh_config.g, a, dh_config.p)
@@ -162,7 +171,7 @@ class SecretChatMethods:
             if my_exchange_id == other_exchange_id:
                 peer.rekeying = [0]
                 return
-        self._log.client[__name__].debug(f'Accepting rekeying secret chat {peer}')
+        self._log.debug(f'Accepting rekeying secret chat {peer}')
         dh_config = await self.get_dh_config()
         random_bytes = os.urandom(256)
         b = int.from_bytes(random_bytes, byteorder="big", signed=False)
@@ -170,16 +179,15 @@ class SecretChatMethods:
         self.check_g_a(g_a, dh_config.p)
         res = pow(g_a, b, dh_config.p)
         auth_key = res.to_bytes(256, 'big', signed=False)
-        key = ChatKey(auth_key)
-        key.fingerprint = struct.unpack('<q', sha1(key.auth_key).digest()[-8:])[0]
-        self.temp_rekeyed_secret_chats[action.exchange_id] = key
+        key_fingerprint = struct.unpack('<q', sha1(auth_key).digest()[-8:])[0]
+        self.temp_rekeyed_secret_chats[action.exchange_id] = auth_key
         peer.rekeying = [2, action.exchange_id]
         g_b = pow(dh_config.g, b, dh_config.p)
         self.check_g_a(g_b, dh_config.p)
         message = DecryptedMessageService(action=DecryptedMessageActionAcceptKey(
             g_b=g_b.to_bytes(256, 'big', signed=False),
             exchange_id=action.exchange_id,
-            key_fingerprint=key.fingerprint
+            key_fingerprint=key_fingerprint
         ))
         message = await self.encrypt_secret_message(peer, message)
         await self.client(SendEncryptedServiceRequest(InputEncryptedChat(peer.id, peer.access_hash), message))
@@ -189,15 +197,14 @@ class SecretChatMethods:
         if peer.rekeying[0] != 1 or not self.temp_rekeyed_secret_chats.get(action.exchange_id, None):
             peer.rekeying = [0]
             return
-        self._log.client[__name__].debug(f'Committing rekeying secret chat {peer}')
+        self._log.debug(f'Committing rekeying secret chat {peer}')
         dh_config = await self.get_dh_config()
         g_b = int.from_bytes(action.g_b, 'big', signed=False)
         self.check_g_a(g_b, dh_config.p)
         res = pow(g_b, self.temp_rekeyed_secret_chats[action.exchange_id], dh_config.p)
         auth_key = res.to_bytes(256, 'big', signed=False)
-        key = ChatKey(auth_key)
-        key.fingerprint = struct.unpack('<q', sha1(key.auth_key).digest()[-8:])[0]
-        if key.fingerprint != action.key_fingerprint:
+        key_fingerprint = struct.unpack('<q', sha1(auth_key).digest()[-8:])[0]
+        if key_fingerprint != action.key_fingerprint:
             message = DecryptedMessageService(action=DecryptedMessageActionAbortKey(
                 exchange_id=action.exchange_id,
             ))
@@ -206,13 +213,13 @@ class SecretChatMethods:
             raise SecurityError("Invalid Key fingerprint")
         message = DecryptedMessageService(action=DecryptedMessageActionCommitKey(
             exchange_id=action.exchange_id,
-            key_fingerprint=key.fingerprint
+            key_fingerprint=key_fingerprint
         ))
         message = await self.encrypt_secret_message(peer, message)
         await self.client(SendEncryptedServiceRequest(InputEncryptedChat(peer.id, peer.access_hash), message))
         del self.temp_rekeyed_secret_chats[action.exchange_id]
         peer.rekeying = [0]
-        peer.key = key
+        peer.auth_key = auth_key
         peer.ttl = 100
         peer.updated = time()
 
@@ -220,7 +227,7 @@ class SecretChatMethods:
         peer = self.get_secret_chat(peer)
         if peer.rekeying[0] != 2 or self.temp_rekeyed_secret_chats.get(action.exchange_id, None):
             return
-        if self.temp_rekeyed_secret_chats.get[action.exchange_id] != action.key_fingerprint:
+        if self.temp_rekeyed_secret_chats.get(action.exchange_id) != action.key_fingerprint:
             message = DecryptedMessageService(action=DecryptedMessageActionAbortKey(
                 exchange_id=action.exchange_id,
             ))
@@ -228,18 +235,18 @@ class SecretChatMethods:
             await self.client(SendEncryptedServiceRequest(InputEncryptedChat(peer.id, peer.access_hash), message))
             raise SecurityError("Invalid Key fingerprint")
 
-        self._log.client[__name__].debug(f'Completing rekeying secret chat {peer}')
+        self._log.debug(f'Completing rekeying secret chat {peer}')
         peer.rekeying = [0]
-        peer.key = self.temp_rekeyed_secret_chats[action.exchange_id]
+        peer.auth_key = self.temp_rekeyed_secret_chats[action.exchange_id]
         peer.ttr = 100
         peer.updated = time()
         del self.temp_rekeyed_secret_chats[action.exchange_id]
         message = DecryptedMessageService(action=DecryptedMessageActionNoop())
         message = await self.encrypt_secret_message(peer, message)
         await self.client(SendEncryptedServiceRequest(InputEncryptedChat(peer.id, peer.access_hash), message))
-        self._log.client[__name__].debug(f'Secret chat {peer} rekeyed succrsfully')
+        self._log.debug(f'Secret chat {peer} rekeyed succrsfully')
 
-    async def handle_decrypted_message(self, decrypted_message, peer: Chats, file):
+    async def handle_decrypted_message(self, decrypted_message, peer: SecretChat, file):
         if isinstance(decrypted_message, (DecryptedMessageService, DecryptedMessageService8)):
             if isinstance(decrypted_message.action, DecryptedMessageActionRequestKey):
                 await self.accept_rekey(peer, decrypted_message.action)
@@ -248,7 +255,7 @@ class SecretChatMethods:
                 await self.commit_rekey(peer, decrypted_message.action)
                 return
             elif isinstance(decrypted_message.action, DecryptedMessageActionCommitKey):
-                await self.commit_rekey(peer, decrypted_message.action)
+                await self.complete_rekey(peer, decrypted_message.action)
                 return
             elif isinstance(decrypted_message.action, DecryptedMessageActionNotifyLayer):
                 peer.layer = decrypted_message.action.layer
@@ -267,7 +274,7 @@ class SecretChatMethods:
                 decrypted_message.action.end_seq_no -= peer.out_seq_no_x
                 decrypted_message.action.start_seq_no //= 2
                 decrypted_message.action.end_seq_no //= 2
-                self._log.client[__name__].warning(f"Resending messages for {peer.id}")
+                self._log.warning(f"Resending messages for {peer.id}")
                 for seq, message in peer.outgoing:
                     if decrypted_message.action.start_seq_no <= seq <= decrypted_message.action.end_seq_no:
                         await self.send_secret_message(peer.id, message.message)
@@ -289,8 +296,8 @@ class SecretChatMethods:
             return await self.handle_decrypted_message(decrypted_message, peer, file)
 
     async def handle_encrypted_update(self, event):
-        if not self.secret_chats.get(event.message.chat_id):
-            self._log.client[__name__].debug("Secret chat not saved. skipping")
+        if not self.session.get_secret_chat_by_id(event.message.chat_id):
+            self._log.debug("Secret chat not saved. skipping")
             return False
         message = event.message
 
@@ -298,8 +305,9 @@ class SecretChatMethods:
 
         auth_key_id = struct.unpack('<q', message.bytes[:8])[0]
         peer = self.get_secret_chat(message.chat_id)
-        if not peer.key.fingerprint or \
-                auth_key_id != peer.key.fingerprint:
+        key_fingerprint = struct.unpack('<q', sha1(peer.auth_key).digest()[-8:])[0]
+
+        if auth_key_id != key_fingerprint:
             await self.close_secret_chat(message.chat_id)
             raise ValueError("Key fingerprint mismatch. Chat closed")
 
@@ -313,7 +321,7 @@ class SecretChatMethods:
                 decrypted_message = self.decrypt_mtproto1(bytes.fromhex(message_key.hex()), message.chat_id,
                                                           bytes.fromhex(encrypted_data.hex()))
                 peer.mtproto = 1
-                self._log.client[__name__].debug(f"Used MTProto 1 with chat {message.chat_id}")
+                self._log.debug(f"Used MTProto 1 with chat {message.chat_id}")
 
         else:
             try:
@@ -324,14 +332,14 @@ class SecretChatMethods:
                 decrypted_message = self.decrypt_mtproto2(bytes.fromhex(message_key.hex()), message.chat_id,
                                                           bytes.fromhex(encrypted_data.hex()))
                 peer.mtproto = 2
-                self._log.client[__name__].debug(f"Used MTProto 2 with chat {message.chat_id}")
+                self._log.debug(f"Used MTProto 2 with chat {message.chat_id}")
         peer.ttr -= 1
         if (peer.ttr <= 0 or (time() - peer.updated) > 7 * 24 * 60 * 60) and peer.rekeying[0] == 0:
             await self.rekey(peer)
         peer.incoming[peer.in_seq_no] = message
         return await self.handle_decrypted_message(decrypted_message, peer, file)
 
-    async def encrypt_secret_message(self, peer, message):
+    async def encrypt_secret_message(self, peer: [int, SecretChat, InputEncryptedChat, EncryptedChat], message):
         peer = self.get_secret_chat(peer)
         peer.ttr -= 1
         if peer.layer > 8:
@@ -354,19 +362,20 @@ class SecretChatMethods:
                 padding += 16
             message += os.urandom(padding)
             is_admin = (0 if peer.admin else 8)
-            first_str = peer.key.auth_key[88 + is_admin:88 + 32 + is_admin]
+            first_str = peer.auth_key[88 + is_admin:88 + 32 + is_admin]
             message_key = sha256(first_str + message).digest()[8:24]
-            aes_key, aes_iv = MTProtoState._calc_key(peer.key.auth_key, message_key,
+            aes_key, aes_iv = MTProtoState._calc_key(peer.auth_key, message_key,
                                                      peer.admin)
         else:
             message_key = sha1(message).digest()[-16:]
-            aes_key, aes_iv = _old_calc_key(peer.key.auth_key, message_key,
+            aes_key, aes_iv = _old_calc_key(peer.auth_key, message_key,
                                             True)
             padding = (16 - len(message) % 16) % 16
             message += os.urandom(padding)
-        message = struct.pack('<q', peer.key.fingerprint) + message_key + AES.encrypt_ige(bytes.fromhex(message.hex()),
-                                                                                          aes_key,
-                                                                                          aes_iv)
+        key_fingerprint = struct.unpack('<q', sha1(peer.auth_key).digest()[-8:])[0]
+        message = struct.pack('<q', key_fingerprint) + message_key + AES.encrypt_ige(bytes.fromhex(message.hex()),
+                                                                                     aes_key,
+                                                                                     aes_iv)
         return message
 
     async def download_secret_media(self, message):
@@ -382,12 +391,13 @@ class SecretChatMethods:
                                                                                                    signed=True)
         if fingerprint != key_fingerprint:
             raise SecurityError("Wrong fingerprint")
-        media = await self.download_file(InputEncryptedFileLocation(message.file.id, message.file.access_hash),
-                                         file=bytes)
-        decrypted_data = AES.decrypt_ige(media, message.media.key, message.media.iv)
-        return decrypted_data
+        media = await self.client.download_file(InputEncryptedFileLocation(message.file.id, message.file.access_hash),
+                                                key=message.media.key,
+                                                iv=message.media.iv)
+        return media
 
-    async def send_secret_message(self, peer_id, message, ttl=0, reply_to_id=None):
+    async def send_secret_message(self, peer_id: [int, SecretChat, InputEncryptedChat, EncryptedChat], message: str,
+                                  ttl: int = 0, reply_to_id: [int, None] = None) -> SentEncryptedMessage:
         peer = self.get_secret_chat(peer_id)
         if peer.layer == 8:
             message = DecryptedMessage8(os.urandom(8), message, DecryptedMessageMediaEmpty())
@@ -408,7 +418,7 @@ class SecretChatMethods:
                                                                                                    byteorder="little",
                                                                                                    signed=True)
 
-        file = await self.upload_file(file, key=key, iv=iv)
+        file = await self.client.upload_file(file, key=key, iv=iv)
         if isinstance(file, InputFileBig):
             file = InputEncryptedFileBigUploaded(file.id, file.parts, fingerprint)
         elif isinstance(file, InputFile):
@@ -416,7 +426,8 @@ class SecretChatMethods:
 
         return file, fingerprint, key, iv
 
-    async def send_secret_document(self, peer, document, thumb: bytes, thumb_w: int, thumb_h: int, file_name: str,
+    async def send_secret_document(self, peer: [int, SecretChat, InputEncryptedChat, EncryptedChat], document,
+                                   thumb: bytes, thumb_w: int, thumb_h: int, file_name: str,
                                    mime_type: str, size: int, attributes=None, ttl=0, caption=""):
         if attributes is None:
             attributes = []
@@ -438,7 +449,8 @@ class SecretChatMethods:
         res = await self.client(SendEncryptedFileRequest(peer.input_chat, data, file=file))
         return res
 
-    async def send_secret_audio(self, peer, audio, duration, mime_type, size, ttl=0, caption=""):
+    async def send_secret_audio(self, peer: [int, SecretChat, InputEncryptedChat, EncryptedChat], audio, duration,
+                                mime_type, size, ttl=0, caption=""):
         peer = self.get_secret_chat(peer)
         file, fingerprint, key, iv = await self.upload_secret_file(audio)
         if peer.layer == 8:
@@ -454,7 +466,8 @@ class SecretChatMethods:
         res = await self.client(SendEncryptedFileRequest(peer.input_chat, data, file=file))
         return res
 
-    async def send_secret_video(self, peer, video, thumb: bytes, thumb_w: int, thumb_h: int, duration: int,
+    async def send_secret_video(self, peer: [int, SecretChat, InputEncryptedChat, EncryptedChat], video, thumb: bytes,
+                                thumb_w: int, thumb_h: int, duration: int,
                                 mime_type: str, w: int,
                                 h: int, size, ttl=0, caption=""):
         peer = self.get_secret_chat(peer)
@@ -478,7 +491,8 @@ class SecretChatMethods:
         res = await self.client(SendEncryptedFileRequest(peer.input_chat, data, file=file))
         return res
 
-    async def send_secret_photo(self, peer, image, thumb, thumb_w, thumb_h, w, h, size, caption="",
+    async def send_secret_photo(self, peer: [int, SecretChat, InputEncryptedChat, EncryptedChat], image, thumb, thumb_w,
+                                thumb_h, w, h, size, caption="",
                                 ttl=0):
         peer = self.get_secret_chat(peer)
 
@@ -498,11 +512,8 @@ class SecretChatMethods:
         res = await self.client(SendEncryptedFileRequest(peer.input_chat, data, file=file))
         return res
 
-    async def notify_layer(self, peer):
-        if isinstance(peer, int):
-            peer = self.secret_chats[peer]
-        else:
-            peer = self.secret_chats[peer.id]
+    async def notify_layer(self, peer: [int, SecretChat, InputEncryptedChat, EncryptedChat]):
+        peer = self.get_secret_chat(peer)
         if peer.layer == 8:
             return
         message = DecryptedMessageService8(action=DecryptedMessageActionNotifyLayer(
@@ -512,31 +523,30 @@ class SecretChatMethods:
             SendEncryptedServiceRequest(peer=InputEncryptedChat(peer.id, peer.access_hash),
                                         data=data))
 
-    async def close_secret_chat(self, peer):
+    async def close_secret_chat(self, peer: [int, SecretChat, InputEncryptedChat, EncryptedChat]):
 
-        if self.secret_chats.get(peer.id, None):
-            del self.secret_chats[peer]
-        if self.temp_secret_chat.get(peer.id, None):
-            del self.temp_secret_chat[peer.id]
+        if self.session.get_secret_chat_by_id(peer.id):
+            self.session.remove_secret_chat_by_id(peer.id, False)
+        if self.session.get_temp_secret_chat_by_id(peer.id):
+            self.session.remove_secret_chat_by_id(peer.id, True)
         try:
             await self.client(DiscardEncryptionRequest(peer.id))
         except EncryptionAlreadyDeclinedError:
-            pass
+            self._log.debug(f"Chat {peer.id} already closed")
 
     def decrypt_mtproto2(self, message_key, chat_id, encrypted_data):
         peer = self.get_secret_chat(chat_id)
-
-        aes_key, aes_iv = MTProtoState._calc_key(self.secret_chats[chat_id].key.auth_key,
+        aes_key, aes_iv = MTProtoState._calc_key(peer.auth_key,
                                                  message_key,
-                                                 not self.secret_chats[chat_id].admin)
+                                                 not peer.admin)
 
         decrypted_data = AES.decrypt_ige(encrypted_data, aes_key, aes_iv)
         message_data_length = struct.unpack('<I', decrypted_data[:4])[0]
         message_data = decrypted_data[4:message_data_length + 4]
         if message_data_length > len(decrypted_data):
             raise SecurityError("message data length is too big")
-        is_admin = peer.admin
-        first_str = peer.key.auth_key[88 + is_admin:88 + 32 + is_admin]
+        is_admin = (8 if peer.admin else 0)
+        first_str = peer.auth_key[88 + is_admin:88 + 32 + is_admin]
 
         if message_key != sha256(first_str + decrypted_data).digest()[8:24]:
             raise SecurityError("Message key mismatch")
@@ -548,12 +558,13 @@ class SecretChatMethods:
         return BinaryReader(message_data).tgread_object()
 
     def decrypt_mtproto1(self, message_key, chat_id, encrypted_data):
-        aes_key, aes_iv = _old_calc_key(self.secret_chats[chat_id].key.auth_key,
+        aes_key, aes_iv = _old_calc_key(self.get_secret_chat(chat_id).auth_key,
                                         message_key,
                                         True)
         decrypted_data = AES.decrypt_ige(encrypted_data, aes_key, aes_iv)
         message_data_length = struct.unpack('<I', decrypted_data[:4])[0]
         message_data = decrypted_data[4:message_data_length + 4]
+
         if message_data_length > len(decrypted_data):
             raise SecurityError("message data length is too big")
 
@@ -576,16 +587,17 @@ class SecretChatMethods:
         self.check_g_a(g_a, dh_config.p)
         res = pow(g_a, b, dh_config.p)
         auth_key = res.to_bytes(256, 'big', signed=False)
-        key = ChatKey(auth_key)
-        key.fingerprint = struct.unpack('<q', sha1(key.auth_key).digest()[-8:])[0]
+
+        key_fingerprint = struct.unpack('<q', sha1(auth_key).digest()[-8:])[0]
         input_peer = InputEncryptedChat(chat_id=chat.id, access_hash=chat.access_hash)
-        secret_chat = Chats(chat.id, chat.access_hash, key, admin=False, user_id=chat.admin_id, input_chat=input_peer)
-        self.secret_chats[chat.id] = secret_chat
+        secret_chat = SecretChat(chat.id, chat.access_hash, auth_key, admin=False, user_id=chat.admin_id,
+                                 input_chat=input_peer)
+        self.session.save_chat(secret_chat, False)
         g_b = pow(dh_config.g, b, dh_config.p)
         self.check_g_a(g_b, dh_config.p)
         result = await self.client(
             AcceptEncryptionRequest(input_peer, g_b=g_b.to_bytes(256, 'big', signed=False),
-                                    key_fingerprint=key.fingerprint))
+                                    key_fingerprint=key_fingerprint))
         await self.notify_layer(chat)
         return result
 
@@ -593,21 +605,15 @@ class SecretChatMethods:
         dh_config = await self.get_dh_config()
         g_a_or_b = int.from_bytes(chat.g_a_or_b, "big", signed=False)
         self.check_g_a(g_a_or_b, dh_config.p)
-        auth_key = pow(g_a_or_b, self.temp_secret_chat[chat.id], dh_config.p).to_bytes(256, "big", signed=False)
-        del self.temp_secret_chat[chat.id]
-        key = ChatKey(auth_key)
-        key.fingerprint = struct.unpack('<q', sha1(key.auth_key).digest()[-8:])[0]
-        if key.fingerprint != chat.key_fingerprint:
+        a = self.session.get_temp_secret_chat_by_id(chat.id).auth_key
+        a = int.from_bytes(a, "big", signed=False)
+        auth_key = pow(g_a_or_b, a, dh_config.p).to_bytes(
+            256, "big", signed=False)
+        self.session.remove_secret_chat_by_id(chat.id, True)
+        key_fingerprint = struct.unpack('<q', sha1(auth_key).digest()[-8:])[0]
+        if key_fingerprint != chat.key_fingerprint:
             raise ValueError("Wrong fingerprint")
-        key.visualization_orig = sha1(key.auth_key).digest()[16:]
-        key.visualization_46 = sha256(key.auth_key).digest()[20:]
         input_peer = InputEncryptedChat(chat_id=chat.id, access_hash=chat.access_hash)
-        self.secret_chats[chat.id] = Chats(
-            chat.id,
-            chat.access_hash,
-            key,
-            True,
-            chat.participant_id,
-            input_peer
-        )
+        self.session.save_chat(SecretChat(chat.id, chat.access_hash, auth_key, True, chat.participant_id,
+                                          input_peer), False)
         await self.notify_layer(chat)
